#include "Arduino.h"
#include "XiaomiDevice.h"


/* XiaomiUDP */
// Read the packet into buffer
int XiaomiUDP::readPacket() {
  int packetSize = _udp.parsePacket();
  if (packetSize) {
      cleanBuffer();
      _udp.read(packetBuffer, packetSize);

      IPAddress remIp = _udp.remoteIP();
      uint16_t remPort = _udp.remotePort();
      Log.verbose("Received packet (%d): '%s' from %d.%d.%d.%d:%d" CR, packetSize, packetBuffer,
                  remIp[0], remIp[1], remIp[2], remIp[3], remPort);
  }

  return packetSize;
}

// Send JSON packet
void XiaomiUDP::sendJSONPacket(JsonObject& cmd, bool multiCast) {
  sendJSONPacket(cmd, _ip, _port, multiCast);
}
void XiaomiUDP::sendJSONPacket(JsonObject& cmd, String strIP, uint16_t port, bool multiCast) {
  IPAddress ip;
  if (!ip.fromString(strIP)) {
    Log.error("Invalid IP address");
    return;
  }
  
  sendJSONPacket(cmd, ip, port, multiCast);
}
void XiaomiUDP::sendJSONPacket(JsonObject& cmd, IPAddress ip, uint16_t port, bool multiCast) {
  #ifndef DISABLE_LOGGING // If logging disabled, skipping this part to save RAM
    char packetData[MAX_PACKET_SIZE];
    cmd.printTo((char*)packetData, cmd.measureLength() + 1);
    Log.verbose("Sending packet(%d): '%s' to %d.%d.%d.%d:%d" CR, cmd.measureLength(), packetData,
                ip[0], ip[1], ip[2], ip[3], port);
  #endif
  
  if (multiCast)
    _udp.beginPacketMulticast(ip, port, WiFi.localIP());
  else
    _udp.beginPacket(ip, port);
  cmd.printTo(_udp);
  _udp.endPacket();
}

// Send string packet
void XiaomiUDP::sendStringPacket(const char *cmd, bool multiCast) {
  sendStringPacket(cmd, _ip, _port, multiCast);
}
void XiaomiUDP::sendStringPacket(const char *cmd, String strIP, uint16_t port, bool multiCast) {
  IPAddress ip;
  if (!ip.fromString(strIP)) {
    Log.error("Invalid IP address");
    return;
  }

  sendStringPacket(cmd, ip, port, multiCast);
}
void XiaomiUDP::sendStringPacket(const char *cmd, IPAddress ip, uint16_t port, bool multiCast) {
  Log.verbose("Sending packet(%d): '%s' to %d.%d.%d.%d:%d" CR, strlen(cmd), cmd,
              ip[0], ip[1], ip[2], ip[3], port);
  
  if (multiCast)
    _udp.beginPacketMulticast(ip, port, WiFi.localIP());
  else
    _udp.beginPacket(ip, port);
  _udp.write(cmd);
  _udp.endPacket();
}

// Clean buffer where to store received data
void XiaomiUDP::cleanBuffer() {
  // Prepare packet buffer
  for (int i = 0; i < MAX_PACKET_SIZE; i++)
    packetBuffer[i] = 0;
}


/* Xiaomi devices */
void XiaomiGenericDevice::updateDevice(String model, String sid, String short_id) {
  this->model = model;
  this->sid = sid;
  this->short_id = short_id;
}
void XiaomiBatteryDevice::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);
  if (data.containsKey("voltage"))
    voltage = data["voltage"].as<String>().toFloat() / 100;
}

void XiaomiGatewayDevice::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);
  if (data.containsKey("rgb")) {
    uint32_t value = data["rgb"];
    brightness = value >> 24;
    rgb = {(uint8_t)(value >> 16 & 0xFF), (uint8_t)(value >> 8 & 0xFF), (uint8_t)(value & 0xFF)};
  }
  
  if (data.containsKey("illumination"))
    illumination = data["illumination"];
}
bool XiaomiGatewayDevice::prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) {
  XiaomiGenericDevice::prepareDataForSendCommand(deviceCommand, data);
  
  // Light
  if (deviceCommand.command == "light") {
    if (deviceCommand.numParams == 4 && isValidNumber(deviceCommand.params[0]) && isValidNumber(deviceCommand.params[1]) &&
        isValidNumber(deviceCommand.params[2]) && isValidNumber(deviceCommand.params[3])) {
      uint8_t reqBrightness = deviceCommand.params[0].toInt();
      RGB reqRGB = {(uint8_t)deviceCommand.params[1].toInt(), (uint8_t)deviceCommand.params[2].toInt(),
                    (uint8_t)deviceCommand.params[3].toInt()};
      data["rgb"] = ((reqRGB.r || reqRGB.g || reqRGB.b)?((uint32_t)reqBrightness << 24L):0) |
                    ((uint32_t)reqRGB.r << 16L) | ((uint32_t)reqRGB.g << 8L) | (uint32_t)reqRGB.b;
      return true;
    }
    
    if (deviceCommand.numParams == 1 && deviceCommand.params[0] == "on") {
      data["rgb"] = ((uint32_t)brightness << 24L) | ((uint32_t)rgb.r << 16L) |
                    ((uint32_t)rgb.g << 8L) | (uint32_t)rgb.b;
     return true;
    }
    
    if (deviceCommand.numParams == 1 && deviceCommand.params[0] == "off") {
      data["rgb"] = 0;
      return true;
    }
    
    if (deviceCommand.numParams == 1 && isValidNumber(deviceCommand.params[0])) {
      uint32_t value = deviceCommand.params[0].toInt();
      uint8_t reqBrightness = value >> 24;
      RGB reqRGB = {(uint8_t)(value >> 16 & 0xFF), (uint8_t)(value >> 8 & 0xFF), (uint8_t)(value & 0xFF)};
      data["rgb"] = ((reqRGB.r || reqRGB.g || reqRGB.b)?((uint32_t)reqBrightness << 24L):0) |
                    ((uint32_t)reqRGB.r << 16L) | ((uint32_t)reqRGB.g << 8L) | (uint32_t)reqRGB.b;
      return true;
    }
  }

  // Play sound
  if (deviceCommand.command == "play") {
    if (deviceCommand.numParams >= 2 && deviceCommand.params[0] == "start" &&
        isValidNumber(deviceCommand.params[1])) {
      data["mid"] = deviceCommand.params[1].toInt();
      if (deviceCommand.numParams >= 3 && isValidNumber(deviceCommand.params[1]))
        data["volume"] = deviceCommand.params[2].toInt();
            
      return true;
    }
     
    if (deviceCommand.numParams >= 1 && deviceCommand.params[0] == "stop") {
       data["mid"] = 10000;
       return true;
    }
  }

  return false;
}

void XiaomiSensorHTDevice::updateData(JsonObject& data) {
  XiaomiBatteryDevice::updateData(data);
  
  if (data.containsKey("temperature"))
    temperature = data["temperature"].as<String>().toFloat() / 100;
  
  if (data.containsKey("humidity"))
    humidity = data["humidity"].as<String>().toFloat() / 100;
}

void XiaomiPlugDevice::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);

  if (data.containsKey("load_voltage"))
    load_voltage = data["load_voltage"].as<String>().toInt() / 100;

  if (data.containsKey("load_power"))
    load_power = data["load_power"].as<String>().toFloat();

  if (data.containsKey("power_consumed"))
    power_consumed = data["power_consumed"].as<String>().toFloat();

  if (data.containsKey("status"))
    status = data["status"].as<String>();
}
bool XiaomiPlugDevice::prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) {
  XiaomiGenericDevice::prepareDataForSendCommand(deviceCommand, data);

  // Status –> on / off
  if (deviceCommand.command == "status" && deviceCommand.numParams == 1 &&
      (deviceCommand.params[0] == "on" || deviceCommand.params[0] == "off")) {
    status = deviceCommand.params[0];
    data["status"] = status;
    return true;
  }

  return false;
}

void Xiaomi86PlugDevice::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);

  if (data.containsKey("status"))
    status = data["status"].as<String>();
}
bool Xiaomi86PlugDevice::prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) {
  XiaomiGenericDevice::prepareDataForSendCommand(deviceCommand, data);

  // Status –> on / off
  if (deviceCommand.command == "status" && deviceCommand.numParams == 1 &&
      (deviceCommand.params[0] == "on" || deviceCommand.params[0] == "off")) {
    status = deviceCommand.params[0];
    data["status"] = status;
    return true;
  }

  return false;
}

void XiaomiCtrlNeutral1Device::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);

  if (data.containsKey("channel_0"))
    channel_0 = data["channel_0"].as<String>();
}
bool XiaomiCtrlNeutral1Device::prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) {
  XiaomiGenericDevice::prepareDataForSendCommand(deviceCommand, data);

  // Status –> on / off
  if (deviceCommand.command == "channel" && deviceCommand.numParams == 2 && deviceCommand.params[0] == "0" &&
      (deviceCommand.params[1] == "on" || deviceCommand.params[1] == "off")) {
    channel_0 = deviceCommand.params[1];
    data["channel_0"] = channel_0;
    return true;
  }

  return false;
}

void XiaomiCtrlNeutral2Device::updateData(JsonObject& data) {
  XiaomiGenericDevice::updateData(data);

  if (data.containsKey("channel_0"))
    channel_0 = data["channel_0"].as<String>();
  if (data.containsKey("channel_1"))
    channel_1 = data["channel_1"].as<String>();
}
bool XiaomiCtrlNeutral2Device::prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) {
  XiaomiGenericDevice::prepareDataForSendCommand(deviceCommand, data);

  // Status –> on / off
  if (deviceCommand.command == "channel" && deviceCommand.numParams == 2) {
    if (deviceCommand.params[0] == "0" && (deviceCommand.params[1] == "on" || deviceCommand.params[1] == "off")) {
      channel_0 = deviceCommand.params[1];
      data["channel_0"] = channel_0;
      return true;
    }
  
    if (deviceCommand.params[0] == "1" && (deviceCommand.params[1] == "on" || deviceCommand.params[1] == "off")) {
      channel_1 = deviceCommand.params[1];
      data["channel_1"] = channel_1;
      return true;
    }
  }

  return false;
}


/* Xiaomi Gateway */
void XiaomiGateway::discoverDevices() {
  Log.notice("Discovering Xiaomi devices for gateway: %s" CR, sid.c_str());

  // Send discovery packet
  StaticJsonBuffer<JSON_OBJECT_SIZE(2) + 64> cmdBuffer; // 64 bytes for strings
  JsonObject& cmd = cmdBuffer.createObject();
  cmd["cmd"] = "get_id_list";
  cmd["sid"] = sid;
  _udpMC->sendJSONPacket(cmd, ip, port);

  // Read response
  unsigned long currentTime = millis();
  while (currentTime + RESPONSE_DELAY > millis()) {
    if (_udpMC->readPacket()) {
      // Parse response packet
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> responseBuffer;
      JsonObject& resp = responseBuffer.parseObject(_udpMC->packetBuffer);
      if (!resp.success()) {
        Log.error("Error when JSON parsing responseBuffer" CR);
        continue;
      }

      // Validate it
      if (resp["cmd"] != "get_id_list_ack" || !resp.containsKey("token") || !resp.containsKey("data")) {
        Log.error("Received unexpected or invalid command" CR);
        continue;
      }

      // Parse data JSON
      String data = resp["data"];
      StaticJsonBuffer<MAX_JSON_ARRAY_SIZE> dataBuffer;
      JsonArray& dataJSON = dataBuffer.parseArray(data);
      if (!dataJSON.success())
        Log.error("Error when JSON parsing data" CR);
      else {
        // Add gateway as a inherited device
        this->addDevice(sid);
        
        // Add all subdevices
        for (JsonArray::iterator it=dataJSON.begin(); it!=dataJSON.end(); ++it) {
          this->addDevice(it->as<String>());
        }
      }

      // Update token
      this->updateKey(resp["token"]);
      break;
    }
  }

  Log.notice("Discovering Xiaomi devices for gateway: %s done" CR, sid.c_str());
}
void XiaomiGateway::updateKey(String token) {
  Log.notice("Updating key for Gateway with IP: %s, received token: %s" CR, ip.c_str(), token.c_str());

  // Prepare data for ecryption
  byte iv[16]  = {0x17, 0x99, 0x6D, 0x09, 0x3D, 0x28, 0xDD, 0xB3, 0xBA, 0x69, 0x5A, 0x2E, 0x6F, 0x58, 0x56, 0x2E};
  byte *pwd = (unsigned char*)password.c_str();
  byte plain[16];
  byte cipher[16];
  strncpy((char *)plain, token.c_str(), 16);

  // Ecnrypt it
  AES aes;
  aes.set_key(pwd, 128);
  aes.cbc_encrypt(plain, cipher, 1, iv);

  // Convert it to HEX string
  sprintf(key, "%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X",
          cipher[0], cipher[1], cipher[2], cipher[3], cipher[4], cipher[5], cipher[6], cipher[7],
          cipher[8], cipher[9], cipher[10], cipher[11], cipher[12], cipher[13], cipher[14], cipher[15]);
  key[32] = '\0';

  Log.notice(F("New key: %s" CR), key);
}

void XiaomiGateway::addDevice(String sid) {
  Log.notice(F("Adding device, SID: %s" CR), sid.c_str());
  if (numDevices >= MAX_DEVICES) {
    Log.warning("Number of devices is breaking MAX_DEVICES constant" CR);
    return;
  }
  if (deviceExists(sid)) {
     Log.notice(F("Device SID: %s already present, skipping" CR), sid.c_str());
     return;
  }

  devices[numDevices].sid = sid;
  devices[numDevices].device = NULL;
  numDevices++;
}
bool XiaomiGateway::deviceExists(String sid) {
  for (int i = 0; i < numDevices; i++)
    if (devices[i].sid == sid)
      return true;

  return false;
}
XiaomiGenericDevice *XiaomiGateway::getDeviceBySID(String sid) {
  for (int i = 0; i < numDevices; i++)
    if (devices[i].sid == sid)
      return devices[i].device;

  return NULL;
}

void XiaomiGateway::readDevice(String sid) {
  int index = -1;
  for (int i = 0; i < numDevices; i++)
    if (devices[i].sid == sid) {
      index = i;
      break;
    }
  
  Log.notice("Reading device: %s" CR, sid.c_str());

  // Send discovery packet
  StaticJsonBuffer<JSON_OBJECT_SIZE(2) + 64> cmdBuffer; // 64 bytes for strings
  JsonObject& cmd = cmdBuffer.createObject();
  cmd["cmd"] = "read";
  cmd["sid"] = sid;
  _udpMC->sendJSONPacket(cmd, ip, port);

  unsigned long currentTime = millis();
  while (currentTime + RESPONSE_DELAY > millis()) {
    if (_udpMC->readPacket()) {
      // Parse response packet
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> responseBuffer;
      JsonObject& resp = responseBuffer.parseObject(_udpMC->packetBuffer);

      if (!resp.success()) {
        Log.error("Error when JSON parsing responseBuffer" CR);
        continue;
      }

      // Validate it
      if (resp["cmd"] != "read_ack" || !resp.containsKey("data")) {
        Log.error("Received unexpected or invalid command" CR);
        continue;
      }

      // Parse data dictionary from packet
      const char *dataJSON = resp["data"];
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> dataBuffer;
      JsonObject& data = dataBuffer.parseObject(dataJSON);
      if (!data.success()) {
        Log.error("Error when JSON parsing data" CR);
        break;
      }

      // Error => not a device
      if (resp.containsKey("error") || data.containsKey("error")) {
        Log.warning("Not a device" CR);
        continue;
      }

      if (resp["sid"] != sid) {
        Log.warning("Received packet for a wrong device" CR);
        continue;
      }

      // Device instance not present, create it depending on the model
      if ((devices[index].device == NULL)) {
        Log.notice("Creating XiaomiDevice instance" CR);
        // Cube
        if (resp["model"] == XiaomiCubeDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiCubeDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Magnet
        else if (resp["model"] == XiaomiGatewayDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiGatewayDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Magnet
        else if (resp["model"] == XiaomiMagnetDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiMagnetDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Motion
        else if (resp["model"] == XiaomiMotionDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiMotionDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Temperature / humidity sensor
        else if (resp["model"] == XiaomiSensorHTDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiSensorHTDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Water leak sensor
        else if (resp["model"] == XiaomiSensorWleakAQ1Device::DEVICE_MODEL)
          devices[index].device = new XiaomiSensorWleakAQ1Device(resp["model"], resp["sid"], resp["short_id"]);
        // Smoke detector
        else if (resp["model"] == XiaomiSmokeDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiSmokeDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Switch
        else if (resp["model"] == XiaomiSwitchDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiSwitchDevice(resp["model"], resp["sid"], resp["short_id"]);
        // 86 Switch 1
        else if (resp["model"] == Xiaomi86SW1Device::DEVICE_MODEL)
          devices[index].device = new Xiaomi86SW1Device(resp["model"], resp["sid"], resp["short_id"]);
        // 86 Switch 2
        else if (resp["model"] == Xiaomi86SW2Device::DEVICE_MODEL)
          devices[index].device = new Xiaomi86SW2Device(resp["model"], resp["sid"], resp["short_id"]);
        // Plug
        else if (resp["model"] == XiaomiPlugDevice::DEVICE_MODEL)
          devices[index].device = new XiaomiPlugDevice(resp["model"], resp["sid"], resp["short_id"]);
        // 86Plug
        else if (resp["model"] == Xiaomi86PlugDevice::DEVICE_MODEL)
          devices[index].device = new Xiaomi86PlugDevice(resp["model"], resp["sid"], resp["short_id"]);
        // Control Neutral 1
        else if (resp["model"] == XiaomiCtrlNeutral1Device::DEVICE_MODEL)
          devices[index].device = new XiaomiCtrlNeutral1Device(resp["model"], resp["sid"], resp["short_id"]);
        // Control Neutral 2
        else if (resp["model"] == XiaomiCtrlNeutral2Device::DEVICE_MODEL)
          devices[index].device = new XiaomiCtrlNeutral2Device(resp["model"], resp["sid"], resp["short_id"]);
        // Uknown device
        else
          devices[index].device = new XiaomiGenericDevice(resp["model"], resp["sid"], resp["short_id"]);
      }
      // Update device info
      else
        Log.notice("Updating XiaomiDevice instance" CR);
        devices[index].device->updateDevice(resp["model"], resp["sid"], resp["short_id"]);

      // Update the device data
      Log.notice("Updating XiaomiDevice data" CR);
      devices[index].device->updateData(data);

      // Call callback for every set in data
      if (_callBack != NULL) {
        for (JsonObject::iterator it=data.begin(); it!=data.end(); ++it)
          (*_callBack)(resp["model"], resp["sid"], it->key, it->value);
      }
      
      break;
    }
  }
  
  Log.notice("Reading device: %s done" CR, devices[index].sid.c_str());
}

void XiaomiGateway::sendCommandToDevice(String sid, XIAOMI_COMMAND deviceCommand) {
  Log.notice("Sending command: %s to device: %s" CR, deviceCommand.command.c_str(), sid.c_str());
  
  XiaomiGenericDevice *device = getDeviceBySID(sid);
  if (device == NULL) {
    Log.error("Uknown device with sid: %s" CR, sid.c_str());
    return;
  }
  
  // Prepare JSON structure
  StaticJsonBuffer<JSON_OBJECT_SIZE(2) + JSON_OBJECT_SIZE(8) + 256> cmdBuffer; // 256 bytes for strings
  JsonObject& cmd = cmdBuffer.createObject();
  cmd["cmd"] = "write";
  cmd["sid"] = sid;
  JsonObject& data = cmd.createNestedObject("data");
  data["key"] = key;

  // Fill in data
  if (!device->prepareDataForSendCommand(deviceCommand, data)) {
    Log.error("Command: %s not recognized" CR, deviceCommand.command.c_str());
    return;
  }

  // Send it
  _udpMC->sendJSONPacket(cmd, ip, port);

  // Read response
  unsigned long currentTime = millis();
  while (currentTime + RESPONSE_DELAY > millis()) {
    if (_udpMC->readPacket()) {
      // Parse response packet
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> responseBuffer;
      JsonObject& resp = responseBuffer.parseObject(_udpMC->packetBuffer);
      if (!resp.success()) {
        Log.error("Error when JSON parsing responseBuffer" CR);
        continue;
      }

      // Validate it
      if (resp["cmd"] != "write_ack" || !resp.containsKey("data")) {
        Log.error("Received unexpected or invalid command" CR);
        continue;
      }

      // Parse data dictionary from packet
      const char *dataJSON = resp["data"];
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> dataBuffer;
      JsonObject& data = dataBuffer.parseObject(dataJSON);
      if (!data.success()) {
        Log.error("Error when JSON parsing data" CR);
        return;
      }
      
      // Update the device data
      Log.notice("Updating XiaomiDevice data" CR);
      device->updateData(data);

      /*
      // Call callback for every set in data
      if (_callBack != NULL) {
        for (JsonObject::iterator it=data.begin(); it!=data.end(); ++it)
          (*_callBack)(resp["model"], resp["sid"], it->key, it->value);
      }
      */

      return;
    }
  }
}


/* Xiaomi Interface */
XiaomiInterface::XiaomiInterface(XIAOMI_GATEWAYS definedGateways[], int numDefinedGateways, XiaomiCallback callBack) {
  Log.notice("Starting XiaomiInterface version: %s" CR, VERSION);
  
  // Prepare and start UDP multicast
  WiFiUDP udpMC;
  udpMC.beginMulticast(WiFi.localIP(), MULTICAST_ADDRESS, MULTICAST_PORT);
  _udpMC = new XiaomiUDP(udpMC, MULTICAST_ADDRESS, MULTICAST_PORT);

  // Init gateway UDP
  WiFiUDP udpServer;
  udpServer.beginMulticast(WiFi.localIP(), MULTICAST_ADDRESS, GATEWAY_DISCOVERY_PORT);
  _udpServer = new XiaomiUDP(udpServer, MULTICAST_ADDRESS, GATEWAY_DISCOVERY_PORT);

  // Copy IP addesses and passwords for gateways
  for (int i = 0; i < MAX_GATEWAYS; i++) {
    _definedGateways[i].ip = (i < numDefinedGateways)?definedGateways[i].ip:"";
    _definedGateways[i].password = (i < numDefinedGateways)?definedGateways[i].password:"";
  }
  
  _callBack = callBack;
  numGateways = 0;
}

void XiaomiInterface::discoverGateways() {
  Log.notice("Searching Xiaomi gateways" CR);

  // Send discovery packet
  _udpServer->sendStringPacket("{\"cmd\":\"whois\"}", true);

  // Wait for 5 seconds to get all responses
  unsigned long currentTime = millis();
  while (currentTime + DISCOVERY_DELAY > millis()) {
    if (_udpServer->readPacket()) {
      // Parse Xiaomi packet
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> jsonBuffer;
      JsonObject& resp = jsonBuffer.parseObject(_udpServer->packetBuffer);
      if (!resp.success()) {
        Log.error("Error when JSON parsing" CR);
        continue;
      }

      // Not expected result
      if (resp["cmd"] != "iam") {
        Log.error("Response does not match return cmd" CR);
        continue;
      }

      // Received packet not from gateway
      if (resp["model"] != "gateway") {
        Log.error("Response must be gateway model" CR);
        continue;
      }

      // Gateway already exists
      if (getGatewayByIPAddress(resp["ip"]) != NULL) {
        Log.notice("Gateway already known, skipping" CR);
        continue;
      }

      // Find password
      String password = "0000000000000000";
      for (int i = 0; i < MAX_GATEWAYS; i++)
        if (_definedGateways[i].ip == resp["ip"]) {
          password = _definedGateways[i].password;
          break;
        }

      // Add gateway
      addGateway(resp["ip"], resp["port"], resp["sid"], password);
    }
  }

  Log.notice("Searching Xiaomi gateways done" CR);
}
void XiaomiInterface::discoverDevices() {
  for (int i = 0; i < numGateways; i++) {
    if (gateways[i] != NULL) {
      processAllPackets();
      gateways[i]->discoverDevices();
    }
  }
}
void XiaomiInterface::readAllDevices() {
  for (int i = 0; i < numGateways; i++) {
    if (gateways[i] != NULL) {
      for (int j = 0; j < gateways[i]->numDevices; j++) {
        processAllPackets();
        gateways[i]->readDevice(gateways[i]->devices[j].sid);
      }
    }
  }
}

bool XiaomiInterface::processPackets() {
  if (_udpMC->readPacket()) {
      // Parse Xiaomi packet
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> packetBuffer;
      JsonObject& resp = packetBuffer.parseObject(_udpMC->packetBuffer);
      if (!resp.success()) {
        Log.error("Error when JSON parsing packet" CR);
        return false;
      }

      if (!resp.containsKey("data")) {
        Log.error("Packet missing data" CR);
        return false;
      }

      // Parse data from Xiaomi packet
      const char *dataJSON = resp["data"];
      StaticJsonBuffer<MAX_JSON_DICT_SIZE> dataBuffer;
      JsonObject& data = dataBuffer.parseObject(dataJSON);
      if (!data.success()) {
        Log.error("Error when JSON parsing data" CR);
        return false;
      }

      // Heartbeat from gateway
      if (resp["cmd"] == "heartbeat" && resp["model"] == "gateway") {
        const char *ip = data["ip"];
        Log.notice("Received gateway heartbeat from IP: %s" CR, ip);
        XiaomiGateway *gateway = getGatewayByIPAddress(ip);
        if (gateway == NULL) {
          Log.error("Unknown gateway with IP address: %s" CR, ip);
          return false;
        }

        // Update gateway key
        gateway->updateKey(resp["token"]);
      }
      // Report for heartbeat from other device
      else if (resp["cmd"] == "report" || resp["cmd"] == "heartbeat") {
        // Update the data
        XiaomiGenericDevice *device = this->getDeviceBySID(resp["sid"]);
        if (device != NULL) {
          Log.notice("Updating XiaomiDevice data" CR);
          device->updateData(data);
        }
        
        // Call callback for every set in data
        if (_callBack != NULL) {
          for (JsonObject::iterator it=data.begin(); it!=data.end(); ++it)
            (*_callBack)(resp["model"], resp["sid"], it->key, it->value);
        }
      }
      else
        Log.error("Unknown packet" CR);
    
    return true;
  }

  return false;
}
void XiaomiInterface::processAllPackets() {
  while (true)
    if (not processPackets())
      return;
}

void XiaomiInterface::sendCommandToDevice(String model, String sid, XIAOMI_COMMAND deviceCommand) {
  XiaomiGenericDevice *device = getDeviceBySID(sid);
  if (device == NULL || device->model != model) {
    Log.error("Command cannot be sent, sid not found or model not matching" CR);
    return;
  }

  XiaomiGateway *gateway = getGatewayForDeviceBySID(sid);
  if (gateway == NULL) {
    Log.error("No matching gateway found" CR);
    return;
  }

  // First process all packets
  processAllPackets();

  // Read command to retrieve device
  if (deviceCommand.command == "read")
    gateway->readDevice(sid);
  else
    gateway->sendCommandToDevice(sid, deviceCommand);
}
XiaomiGenericDevice *XiaomiInterface::getDeviceBySID(String sid) {
  for (int i = 0; i < numGateways; i++)
    if (gateways[i] != NULL && gateways[i]->deviceExists(sid))
      return gateways[i]->getDeviceBySID(sid);

  return NULL;
}
XiaomiGateway *XiaomiInterface::getGatewayForDeviceBySID(String sid) {
  for (int i = 0; i < numGateways; i++)
    if (gateways[i] != NULL && gateways[i]->deviceExists(sid))
      return gateways[i];

  return NULL;
}
XiaomiGateway *XiaomiInterface::getGatewayByIPAddress(String ip) {
  for (int i = 0; i < numGateways; i++)
    if (gateways[i] != NULL && gateways[i]->ip == ip)
      return gateways[i];

  return NULL;
}
void XiaomiInterface::addGateway(String ip, uint16_t port, String sid, String key) {
  if (numGateways >= MAX_GATEWAYS) {
    Log.warning("Number of gateways is breaking MAX_GATEWAYS constant" CR);
    return;
  }

  Log.notice(F("Adding gateway, IP: %s, Port: %d, SID: %s, Key: %s" CR), ip.c_str(), port, sid.c_str(), key.c_str());
  gateways[numGateways] = new XiaomiGateway(ip, port, sid, key, _udpMC, _callBack);
  numGateways++;
}


/* Helper functions */
boolean isValidNumber(String str) {
  boolean isNum = false;
  if (!(str.charAt(0) == '+' || str.charAt(0) == '-' || isDigit(str.charAt(0))))
    return false;

  for (byte i=1; i < str.length(); i++)
    if (!(isDigit(str.charAt(i)) || str.charAt(i) == '.'))
      return false;
   
  return true;
}
