/**************************************************************************/
/*! 
    @file     XiaomiReader.ino
    @version  1.0.4
    @author   Jiri Jasa
    @company  DESO development, s.r.o.

    This tool is provides interface between Xiaomi Smart Home Gateway and Loxone
*/
/**************************************************************************/

#include <ESP8266WiFi.h>
#include <WiFiUDP.h>
#include <ArduinoLog.h>

#include "XiaomiDevice.h"


/* Define your setup here */

// Wifi config
char ssid[] = "MyWiFi";        // Fill in your network SSID
char password[] = "12345678";  // Fill in your network password 

// Loxone IP address + port definition
/* You can leave next line commented out to sent packets to broadcast
 * If you don't have Loxone and ESP device on the same subnet make sure to
 * uncomment following line and fill in correct IP address
 */
//#define LOXONE_IP IPAddress(192, 168, 1, 254)  // IP address of your Loxone, if you don't know it
                                                 // and your Loxone is sharing same subnet with ESP device
                                                 // keep this line commented out
const unsigned int LOXONE_SEND_PORT = 55556;     // Port to send Loxone UDP commands
const unsigned int LOXONE_RECEIVE_PORT = 44445;  // Port to listen for Loxone UDP commands

// Definition of Xiaomi gateways
XIAOMI_GATEWAYS GATEWAYS[] = {{"192.168.3.100", "AAAAAAAAAAAAAAAA"},
                              {"192.168.3.101", "BBBBBBBBBBBBBBBB"},
                              };

// Definition of sockets
WiFiUDP udpLoxoneSender;
WiFiUDP udpLoxoneReceiver;

// XiaomiDevices interface
XiaomiInterface *xiaomi;

// Define max sizes + buffers
#define MAX_LOXONE_PACKET_SIZE 128
char loxoneBuffer[MAX_LOXONE_PACKET_SIZE];

// Definition of callback + reading function
void xiaomiCallBack(String model, String sid, String key, String value);
void readLoxoneCommand();

// Helper functions
void cleanLoxoneBuffer();
String getValue(String data, char separator, int index);


void setup() {
  //Init Serial
  Serial.begin(115200);
  Serial.println("");

  //Init logging - think about disabling logging during production
  Log.begin(LOG_LEVEL_VERBOSE, &Serial);

  // MAC address info
  byte mac[6];
  WiFi.macAddress(mac);
  Serial.print("WiFi adapter MAC address: ");
  Serial.print(mac[5], HEX); Serial.print(":"); Serial.print(mac[4], HEX); Serial.print(":");
  Serial.print(mac[3], HEX); Serial.print(":"); Serial.print(mac[2], HEX); Serial.print(":");
  Serial.print(mac[1], HEX); Serial.print(":"); Serial.print(mac[0], HEX); Serial.println(); 
  
  // Connect to Wifi
  Serial.print("Connecting to "); Serial.println(ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.println("WiFi connected");

  // Print wifi info
  Serial.print("SSID: "); Serial.println(WiFi.SSID());
  IPAddress ip = WiFi.localIP();
  Serial.print("IP Address: "); Serial.println(ip);
  long rssi = WiFi.RSSI();
  Serial.print("Signal strength (RSSI): "); Serial.print(rssi); Serial.println(" dBm");

  // Prepare Loxone UDP
  #ifdef LOXONE_IP
    Serial.print("Loxone interface to send UDP packets: "); Serial.print(LOXONE_IP); Serial.print(":"); Serial.println(LOXONE_SEND_PORT);
  #else
    Serial.print("Loxone interface to send UDP is using broadcast (255.255.255.255), port: "); Serial.println(LOXONE_SEND_PORT);
  #endif
  udpLoxoneSender.begin(LOXONE_SEND_PORT);
  
  Serial.print("Loxone interface to receive UDP packets, port: "); Serial.println(LOXONE_RECEIVE_PORT);
  udpLoxoneReceiver.begin(LOXONE_RECEIVE_PORT);

  // Initialize Xiaomi interface
  xiaomi = new XiaomiInterface(GATEWAYS, sizeof(GATEWAYS) / sizeof(XIAOMI_GATEWAYS), &xiaomiCallBack);
  xiaomi->discoverGateways();
  xiaomi->discoverDevices();
  xiaomi->readAllDevices();
}

void loop() {
  // Process Xiaomi packets
  xiaomi->processPackets();
  // Process Loxone packets
  readLoxoneCommand();
}


/* Sending the data */

// Callback to handle data update
void xiaomiCallBack(String model, String sid, String key, String value) {
  Log.notice("Received Xiaomi callback, model: %s, sid: %s, key: %s, value: %s" CR, model.c_str(), sid.c_str(), key.c_str(), value.c_str());
  String packet  = model + " " + sid + " " + key + " " + value;

  // Loxone IP defined -> unicast
  #ifdef LOXONE_IP
    Log.notice(("Sending packet to Loxone (%d): '%s' to %d.%d.%d.%d:%d" CR), packet.length(), packet.c_str(),
                LOXONE_IP[0], LOXONE_IP[1], LOXONE_IP[2], LOXONE_IP[3], LOXONE_SEND_PORT);
    udpLoxoneSender.beginPacket(LOXONE_IP, LOXONE_SEND_PORT);
  // IP not set -> broadcast
  #else
    IPAddress brIP = ~WiFi.subnetMask() | WiFi.gatewayIP();
    Log.notice(("Sending packet to Loxone (%d): '%s' over broadcast: %d.%d.%d.%d using port: %d" CR),
               packet.length(), packet.c_str(), brIP[0], brIP[1], brIP[2], brIP[3], LOXONE_SEND_PORT);
    udpLoxoneSender.beginPacketMulticast(brIP, LOXONE_SEND_PORT, WiFi.localIP());
  #endif
  udpLoxoneSender.write(packet.c_str());
  udpLoxoneSender.endPacket();
}


/* Receiving the data */

// Read Loxone command
void readLoxoneCommand() {
  int packetSize = udpLoxoneReceiver.parsePacket();
  if (packetSize) {
    // Read packet
    cleanLoxoneBuffer();
    udpLoxoneReceiver.read(loxoneBuffer, packetSize);
    IPAddress remIp = udpLoxoneReceiver.remoteIP();
    uint16_t remPort = udpLoxoneReceiver.remotePort();
    Log.verbose("Received packet (%d): '%s' from %d.%d.%d.%d:%d" CR, packetSize, loxoneBuffer,
                remIp[0], remIp[1], remIp[2], remIp[3], remPort);

    // Parse command
    XIAOMI_COMMAND deviceCommand;
    String str = String(loxoneBuffer);
    String model = getValue(str, ' ', 0);
    String sid = getValue(str, ' ', 1);
    deviceCommand.command = getValue(str, ' ', 2);
    deviceCommand.numParams = 0;
     
    // Fill int params
    int prevIndex = 0, count = 0;
    while (prevIndex >= 0) {
      prevIndex = str.indexOf(' ', prevIndex) + 1; // Find next space
      if (prevIndex == 0)
         break;
  
      count++;
      if (count > 2) {
        deviceCommand.params[deviceCommand.numParams] = getValue(str, ' ', count);
        deviceCommand.numParams++;
      }
    }

    // Send command to device
    xiaomi->sendCommandToDevice(model, sid, deviceCommand);
  }
}

/* Helper functions */

// Clear buffer
void cleanLoxoneBuffer() {
  for (int i = 0; i < MAX_LOXONE_PACKET_SIZE; i++)
    loxoneBuffer[i] = 0;
}

// String splitting by delimiter
String getValue(String data, char separator, int index) {
  int found = 0;
  int strIndex[] = { 0, -1 };
  int maxIndex = data.length() - 1;
  for (int i = 0; i <= maxIndex && found <= index; i++) {
    if (data.charAt(i) == separator || i == maxIndex) {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i+1 : i;
    }
  }

  return (found > index) ? data.substring(strIndex[0], strIndex[1]) : "";
}
